function uu = OMNI_apply_HSS_fsym(NODES,qq)

nboxes = size(NODES,2);
FIELDS = cell(2,nboxes);

% Initialize the fields on the leaves
%  - Construct the outgoing fields.
%  - Set the incoming fields to zero.
for ibox = nboxes:(-1):2
  if (NODES{5,ibox}==0)
    ind            = NODES{6,ibox} - 1 + (1:NODES{7,ibox});
    FIELDS{1,ibox} = OMNI_skel_proj(NODES{30,ibox},NODES{31,ibox},qq(ind,:));
    FIELDS{2,ibox} = zeros(length(NODES{20,ibox}),size(qq,2));
  end
end

% Construct all outgoing potentials:
for ibox = nboxes:(-1):2
  if (NODES{5,ibox}>0) % ibox has at least one child.
    if (NODES{5,ibox}==2) % ibox has two children.
      ison1 = NODES{4,ibox}(1);
      ison2 = NODES{4,ibox}(2);
      qloc  = [FIELDS{1,ison1};FIELDS{1,ison2}];
    elseif (NODES{5,ibox}==1) % ibox has one child.
      ison  = NODES{4,ibox}(1);
      qloc  = FIELDS{1,ison};
    end
    FIELDS{1,ibox} = OMNI_skel_proj(NODES{30,ibox},NODES{31,ibox},qloc);
  end
end

% Construct all incoming potentials:
FIELDS{2,2} = NODES{46,2} *FIELDS{1,3};
FIELDS{2,3} = NODES{46,2}'*FIELDS{1,2};
for ibox = 2:nboxes
  if (NODES{5,ibox}==2) % ibox has two sons
    ison1   = NODES{4,ibox}(1);
    ison2   = NODES{4,ibox}(2);
    n1      = length(NODES{20,ison1});
    n2      = length(NODES{20,ison2});
    u_long  = OMNI_skel_eval(NODES{30,ibox},NODES{31,ibox},FIELDS{2,ibox});
    FIELDS{2,ison1} = u_long(1:n1,:)        + NODES{46,ison1} *FIELDS{1,ison2};
    FIELDS{2,ison2} = u_long(n1 + (1:n2),:) + NODES{46,ison1}'*FIELDS{1,ison1};
  elseif (NODES{5,ibox}==1) % ibox has a one son
    ison            = NODES{4,ibox};
    FIELDS{2,ison}  = OMNI_skel_eval(NODES{30,ibox},NODES{31,ibox},FIELDS{2,ibox});
  end
end

% Construct the potential on all the leaves:
uu = zeros(size(qq));
for ibox = nboxes:(-1):2
  if (NODES{5,ibox}==0)
    ind       = NODES{6,ibox} - 1 + (1:NODES{7,ibox});
    uu(ind,:) = OMNI_skel_eval(NODES{30,ibox},NODES{31,ibox},FIELDS{2,ibox}) + ...
                NODES{40,ibox}*qq(ind,:);
  end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function qq_proj = OMNI_skel_proj(T,J,qq)

if (min(size(T)) == 0)
  qq_proj = qq(J,:);
else
  k       = size(T,1);
  qq_proj = qq(J(1:k),:) + T*qq(J((k+1):end),:);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function uu_eval = OMNI_skel_eval(T,J,uu)

uu_eval = zeros(length(J),size(uu,2));
if (min(size(T)) == 0)
  uu_eval(J,:) = uu;
else
  k                       = size(T,1);
  uu_eval(J(1:k),:)       = uu;
  uu_eval(J((k+1):end),:) = T'*uu;
end
