% This file was created in preparation for the Montestigliano workshop.
%
% The immediate objective is to set up multibody CEs.

function main27

clear
rng(0)

flag_test = 7;

if (sum(flag_test == 1) > 0)
  LOCAL_test_exterior_singlebody()
end

if (sum(flag_test == 2) > 0)
  LOCAL_test_interior_singlebody()
end

if (sum(flag_test == 3) > 0)
  LOCAL_test_exterior_twoCEs()
end

if (sum(flag_test == 4) > 0)
  LOCAL_test_between_CEs()
end

if (sum(flag_test == 5) > 0)
  LOCAL_test_submatrix_functionality()
end

if (sum(flag_test == 6) > 0)
  LOCAL_test_tear()
end

if (sum(flag_test == 7) > 0)
  LOCAL_test_pacman()
end


return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function LOCAL_test_exterior_singlebody()

fprintf(1,'Executing a test for an exterior problems for a single CE.\n')
Npan      = 15;
Ngau      = 10;
[ww,C,xx_int,xx_ext] = LOCAL_geo_elli(Npan,Ngau);
ntot      = size(C,2);
zz        = [0;0];
CE        = [C;...
             zz(1)*ones(1,ntot);...
             zz(2)*ones(1,ntot);...
             -ones(1,ntot);...
             ww];
A         = LOCAL_get_A_diag(CE,1:ntot);
qq_int    = randn(size(xx_int,2),1);
u_dir     = LOCAL_pot_from_charges(CE([1,4],:),xx_int,qq_int);
u_ref     = LOCAL_pot_from_charges(xx_ext,xx_int,qq_int);
sww       = sqrt(ww);
sigma     = (1./sww').*(A\(sww'.*u_dir));
u_app     = LOCAL_eval_pot(xx_ext,CE,sigma);
fprintf(1,'   ||u_ref - u_app|| / ||u_app|| = %18.10e\n',...
        norm(u_ref - u_app)/norm(u_app))

figure(1)
plot(CE(1,:),CE(4,:),'.',...
     zz(1),zz(2),'o',...
     xx_int(1,:),xx_int(2,:),'rx',...
     xx_ext(1,:),xx_ext(2,:),'bx')
legend('C','center','xxint','xxext')
axis equal

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function LOCAL_test_interior_singlebody()

fprintf(1,'Testing for an interior problem (single CE).\n')
Npan      = 15;
Ngau      = 10;
[ww0,C0,xx_int,xx_ext] = LOCAL_geo_elli(Npan,Ngau);
ntot = size(C0,2);
zz      = [2;2];
CE = [C0;...
           zz(1)*ones(1,ntot);...
           zz(2)*ones(1,ntot);...
           ones(1,ntot);...
           ww0];
A       = LOCAL_get_A_diag(CE,1:ntot);
qq_ext  = randn(size(xx_ext,2),1);
u_dir   = LOCAL_pot_from_charges(C0([1,4],:),xx_ext,qq_ext);
u_ref   = LOCAL_pot_from_charges(xx_int,xx_ext,qq_ext);
sww     = sqrt(ww0');
sigma   = (1./sww).*(A\(sww.*u_dir));
u_app   = LOCAL_eval_pot(xx_int,CE,sigma);
fprintf(1,'   ||u_ref - u_app|| / ||u_app|| = %18.10e\n',...
        norm(u_ref - u_app)/norm(u_app))

figure(1)
plot(CE(1,:),CE(4,:),'.',...
     zz(1),zz(2),'o',...
     xx_int(1,:),xx_int(2,:),'bx',...
     xx_ext(1,:),xx_ext(2,:),'rx')
legend('C','center','xxint','xxext')
axis equal

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function   LOCAL_test_exterior_twoCEs()

fprintf(1,'Test for a domain exterior to TWO CEs\n')
Npan      = 15;
Ngau      = 10;
[ww0,C0,~,~] = LOCAL_geo_elli(Npan,Ngau);
xx_int  = [-1,0,1,4;...
           0.2,-0.3,0.4,1];
xx_ext  = [-3,-2,-1;...
            2, 3, 4];
zz      = [0,5;...
           0,1];
C1      = C0;
C2      = C0;
C2(1,:) = C2(1,:) + 5;
C2(4,:) = C2(4,:) + 1;
C       = [C1,C2];
ww      = [ww0,ww0];
ntot    = size(C,2);
n1      = size(C0,2);
CE = [C;...
           [zz(1,1)*ones(1,n1),zz(1,2)*ones(1,n1)];...
           [zz(2,1)*ones(1,n1),zz(2,2)*ones(1,n1)];...
           -ones(1,2*n1);...
           ww];
A       = LOCAL_get_A_diag(CE,1:ntot);
qq_int  = randn(size(xx_int,2),1);
u_dir   = LOCAL_pot_from_charges(C([1,4],:),xx_int,qq_int);
u_ref   = LOCAL_pot_from_charges(xx_ext,xx_int,qq_int);
sww     = sqrt(ww');
sigma   = (1./sww).*(A\(sww.*u_dir));
u_app   = LOCAL_eval_pot(xx_ext,CE,sigma);
fprintf(1,'   ||u_ref - u_app|| / ||u_app|| = %18.10e\n',...
        norm(u_ref - u_app)/norm(u_app))

figure(1)
plot(CE(1,:),CE(4,:),'.',...
     zz(1,:),zz(2,:),'o',...
     xx_int(1,:),xx_int(2,:),'rx',...
     xx_ext(1,:),xx_ext(2,:),'bx')
legend('C','centers','xxint','xxext')
axis equal

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function LOCAL_test_between_CEs()

fprintf(1,'Test for a domain BETWEEN two CEs\n')
Npan      = 15;
Ngau      = 10;
[ww0,C0,~,~] = LOCAL_geo_elli(Npan,Ngau);
xx_src  = [-1,0,1,5;...
           0.2,-0.3,0.4,2];
xx_trg  = [ 3, 2,-2;...
            0, 1.4, 1.3];
zz      = [5,0;...
           5,0];
C1      = 2*C0;
C2      = C0;
C       = [C1,C2];
ww      = [2*ww0,ww0];
ntot    = size(C,2);
n1      = size(C0,2);
CE = [C;...
           [zz(1,1)*ones(1,n1),zz(1,2)*ones(1,n1)];...
           [zz(2,1)*ones(1,n1),zz(2,2)*ones(1,n1)];...
           ones(1,n1),-ones(1,n1);...
           ww];
A       = LOCAL_get_A_diag(CE,1:ntot);
qq_src  = randn(size(xx_src,2),1);
u_dir   = LOCAL_pot_from_charges(C([1,4],:),xx_src,qq_src);
u_ref   = LOCAL_pot_from_charges(xx_trg,xx_src,qq_src);
sww     = sqrt(ww');
sigma   = (1./sww).*(A\(sww.*u_dir));
u_app   = LOCAL_eval_pot(xx_trg,CE,sigma);
fprintf(1,'   ||u_ref - u_app|| / ||u_app|| = %18.10e\n',...
        norm(u_ref - u_app)/norm(u_app))

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function LOCAL_test_submatrix_functionality()

fprintf(1,'Test for a domain BETWEEN two CEs\n')
Npan      = 15;
Ngau      = 10;
[ww0,C0,~,~] = LOCAL_geo_elli(Npan,Ngau);
zz      = [5,0;...
           5,0];
C1      = 2*C0;
C2      = C0;
C       = [C1,C2];
ww      = [2*ww0,ww0];
ntot    = size(C,2);
n1      = size(C0,2);
CE = [C;...
           [zz(1,1)*ones(1,n1),zz(1,2)*ones(1,n1)];...
           [zz(2,1)*ones(1,n1),zz(2,2)*ones(1,n1)];...
           ones(1,n1),-ones(1,n1);...
           ww];
A       = LOCAL_get_A_diag(CE,1:ntot);

ind1 = 11:20;
A11 = LOCAL_get_A_diag(CE,ind1);
fprintf(1,'Error in A11 = %12.5e\n',max(max(abs(A(ind1,ind1) - A11))))

ind1 = [3,7,3,11];
A11 = LOCAL_get_A_diag(CE,ind1);
fprintf(1,'Error in A11 = %12.5e\n',max(max(abs(A(ind1,ind1) - A11))))

ind1 = [3,7,3,11,ntot-5];
ind2 = [4:6,ntot-3];
A12 = LOCAL_get_A_offd(CE,ind1,ind2);
fprintf(1,'Error in A12 = %12.5e\n',max(max(abs(A(ind1,ind2) - A12))))

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function LOCAL_test_tear()

fprintf(1,'Testing a domain with a corner.\n')
Npan      = 10;
Ngau      = 10;
Nref      = 5;
[ww0,C0,xx_int,xx_ext,tt] = LOCAL_geo_tear(Npan,Ngau,Nref);
ntot = size(C0,2);
figure(1)
plot(C0(1,:),C0(4,:),'.',...
     xx_int(1,:),xx_int(2,:),'bx',...
     xx_ext(1,:),xx_ext(2,:),'rx')
legend('C','xxint','xxext')
axis equal

fprintf(1,'Testing an interior problem.\n')
zz      = [2;2];
CE      = [C0;...
           zz(1)*ones(1,ntot);...
           zz(2)*ones(1,ntot);...
           ones(1,ntot);...
           ww0];
A       = LOCAL_get_A_diag(CE,1:ntot);
qq_ext  = randn(size(xx_ext,2),1);
u_dir   = LOCAL_pot_from_charges(C0([1,4],:),xx_ext,qq_ext);
u_ref   = LOCAL_pot_from_charges(xx_int,xx_ext,qq_ext);
sww     = sqrt(ww0');
sigma   = (1./sww).*(A\(sww.*u_dir));
u_app   = LOCAL_eval_pot(xx_int,CE,sigma);
fprintf(1,'   ||u_ref - u_app|| / ||u_app|| = %18.10e\n',...
        norm(u_ref - u_app)/norm(u_app))
figure(2)
plot(tt,sigma)
title('The charge "sigma" for the interior problem.')

fprintf(1,'Testing an exterior problem.\n')
zz      = [1;0];
CE      = [C0;...
           zz(1)*ones(1,ntot);...
           zz(2)*ones(1,ntot);...
           -ones(1,ntot);...
           ww0];
A       = LOCAL_get_A_diag(CE,1:ntot);
qq_int  = randn(size(xx_int,2),1);
u_dir   = LOCAL_pot_from_charges(C0([1,4],:),xx_int,qq_int);
u_ref   = LOCAL_pot_from_charges(xx_ext,xx_int,qq_int);
sww     = sqrt(ww0');
sigma   = (1./sww).*(A\(sww.*u_dir));
u_app   = LOCAL_eval_pot(xx_ext,CE,sigma);
fprintf(1,'   ||u_ref - u_app|| / ||u_app|| = %18.10e\n',...
        norm(u_ref - u_app)/norm(u_app))
figure(3)
plot(tt,sigma)
title('The charge "sigma" for the exterior problem.')

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function LOCAL_test_pacman()

fprintf(1,'Testing a domain with a re-entrant corner.\n')
Npan      = 10;
Ngau      = 10;
Nref      = 25;
[ww0,C0,xx_int,xx_ext,tt] = LOCAL_geo_pacman(Npan,Ngau,Nref);
ntot = size(C0,2);
figure(1)
plot(C0(1,:),C0(4,:),'.',...
     xx_int(1,:),xx_int(2,:),'bx',...
     xx_ext(1,:),xx_ext(2,:),'rx')
legend('C','xxint','xxext')
axis equal

fprintf(1,'Testing an interior problem.\n')
zz      = [2;2];
CE      = [C0;...
           zz(1)*ones(1,ntot);...
           zz(2)*ones(1,ntot);...
           ones(1,ntot);...
           ww0];
A       = LOCAL_get_A_diag(CE,1:ntot);
qq_ext  = randn(size(xx_ext,2),1);
u_dir   = LOCAL_pot_from_charges(C0([1,4],:),xx_ext,qq_ext);
u_ref   = LOCAL_pot_from_charges(xx_int,xx_ext,qq_ext);
sww     = sqrt(ww0');
sigma   = (1./sww).*(A\(sww.*u_dir));
u_app   = LOCAL_eval_pot(xx_int,CE,sigma);
fprintf(1,'   ||u_ref - u_app|| / ||u_app|| = %18.10e\n',...
        norm(u_ref - u_app)/norm(u_app))
figure(2)
plot(tt,sigma)
title('The charge "sigma" for the interior problem.')

fprintf(1,'Testing an exterior problem.\n')
zz      = [1;0];
CE      = [C0;...
           zz(1)*ones(1,ntot);...
           zz(2)*ones(1,ntot);...
           -ones(1,ntot);...
           ww0];
A       = LOCAL_get_A_diag(CE,1:ntot);
qq_int  = randn(size(xx_int,2),1);
u_dir   = LOCAL_pot_from_charges(C0([1,4],:),xx_int,qq_int);
u_ref   = LOCAL_pot_from_charges(xx_ext,xx_int,qq_int);
sww     = sqrt(ww0');
sigma   = (1./sww).*(A\(sww.*u_dir));
u_app   = LOCAL_eval_pot(xx_ext,CE,sigma);
fprintf(1,'   ||u_ref - u_app|| / ||u_app|| = %18.10e\n',...
        norm(u_ref - u_app)/norm(u_app))
figure(3)
plot(tt,sigma)
title('The charge "sigma" for the exterior problem.')

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function D = LOCAL_get_double_kernel(C)

N = size(C,2);

dd1 = C(1,:)' * ones(1,N) - ones(N,1) * C(1,:);
dd2 = C(4,:)' * ones(1,N) - ones(N,1) * C(4,:);
nn1 = ones(N,1)*( C(5,:)./sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:)));
nn2 = ones(N,1)*(-C(2,:)./sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:)));

ddsq = (dd1.*dd1 + dd2.*dd2) + eye(N);

D = -(1/(2*pi))*(nn1.*dd1 + nn2.*dd2)./ddsq;

d_vec = -(-C(6,:).*C(2,:) + C(3,:).*C(5,:))./((4*pi)*(C(2,:).^2 + C(5,:).^2).^1.5);

D = D + diag(d_vec);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function u = LOCAL_eval_double(xx_trg,C,ww,sigma)

beta  = -1/(2*pi);
n_trg = size(xx_trg,2);
n_src = size(C,2);

dd1 = xx_trg(1,:)' * ones(1,n_src) - ones(n_trg,1) * C(1,:);
dd2 = xx_trg(2,:)' * ones(1,n_src) - ones(n_trg,1) * C(4,:);
nn1 = ones(n_trg,1)*( C(5,:)./sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:)));
nn2 = ones(n_trg,1)*(-C(2,:)./sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:)));

D = beta*(nn1.*dd1 + nn2.*dd2)./(dd1.*dd1 + dd2.*dd2);

u = D*(ww' .* sigma);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [tt_panels,tt,ww,ipanel,C] = LOCAL_get_geom(Npan,Ngau,flag_cont)

if strcmp(flag_cont,'elli')

  r1            = 2;    % Parameter controlling the discretization.
  r2            = 1.5;  % Parameter controlling the discretization. 
  t_sta         = 0;    % Start point of the discretization.
  t_end         = 2*pi; % End   point of the discretization. 
  tt_panels     = linspace(t_sta,t_end,Npan+1);
  [tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
  C             = zeros(6,length(tt));
  C(1,:)        =  r1*cos(tt);
  C(2,:)        = -r1*sin(tt);
  C(3,:)        = -r1*cos(tt);
  C(4,:)        =  r2*sin(tt);
  C(5,:)        =  r2*cos(tt);
  C(6,:)        = -r2*sin(tt);
  ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
  h             = 1/Ngau;
  ipanel        = floor(linspace(1+0.5*h,Npan+1-0.5*h,Npan*Ngau));

elseif strcmp(flag_cont,'star')
    
  r             = 0.2;  % Parameter controlling the discretization.
  k             = 5;    % Parameter controlling the discretization. 
  t_sta         = 0;    % Start point of the discretization.
  t_end         = 2*pi; % End   point of the discretization.
  tt_panels     = linspace(t_sta,t_end,Npan+1);
  [tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
  C             = zeros(6,Npan*Ngau);
  C(1,:)        =   1.5*cos(tt) + (r/2)*            cos((k+1)*tt) + (r/2)*            cos((k-1)*tt);
  C(2,:)        = - 1.5*sin(tt) - (r/2)*(k+1)*      sin((k+1)*tt) - (r/2)*(k-1)*      sin((k-1)*tt);
  C(3,:)        = - 1.5*cos(tt) - (r/2)*(k+1)*(k+1)*cos((k+1)*tt) - (r/2)*(k-1)*(k-1)*cos((k-1)*tt);
  C(4,:)        =       sin(tt) + (r/2)*            sin((k+1)*tt) - (r/2)*            sin((k-1)*tt);
  C(5,:)        =       cos(tt) + (r/2)*(k+1)*      cos((k+1)*tt) - (r/2)*(k-1)*      cos((k-1)*tt);
  C(6,:)        = -     sin(tt) - (r/2)*(k+1)*(k+1)*sin((k+1)*tt) + (r/2)*(k-1)*(k-1)*sin((k-1)*tt);
  ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
  h             = 1/Ngau;
  ipanel        = floor(linspace(1+0.5*h,Npan+1-0.5*h,Npan*Ngau));
  
elseif strcmp(flag_cont,'tear')
    
  t_sta         = -pi;    % Start point of the discretization.
  t_end         =  pi;    % End   point of the discretization.
  tt_panels     = linspace(t_sta,t_end,Npan+1);
  [tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
  C             = zeros(6,Npan*Ngau);
  C(1,:)        =      2*sign(tt).*sin(tt/2);
  C(2,:)        =        sign(tt).*cos(tt/2);
  C(3,:)        = -(1/2)*sign(tt).*sin(tt/2);
  C(4,:)        = -tan(pi/4)*sin(tt);
  C(5,:)        = -tan(pi/4)*cos(tt);
  C(6,:)        =  tan(pi/4)*sin(tt);
  ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
  h             = 1/Ngau;
  ipanel        = floor(linspace(1+0.5*h,Npan+1-0.5*h,Npan*Ngau));
  
elseif strcomp(flag_cont,'pacm')
    
  t_sta         = -pi;    % Start point of the discretization.
  t_end         =  pi;    % End   point of the discretization.
  tt_panels     = linspace(t_sta,t_end,Npan+1);
  [tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
  C             = zeros(6,Npan*Ngau);
  C(1,:)        = -     sign(tt).*sin(1.5*tt)+1;
  C(2,:)        = - 1.5*sign(tt).*cos(1.5*tt);
  C(3,:)        =  2.25*sign(tt).*sin(1.5*tt);
  C(4,:)        =     tan(3*pi/4)*sin(tt);
  C(5,:)        =     tan(3*pi/4)*cos(tt);
  C(6,:)        = -   tan(3*pi/4)*sin(tt);
  ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
  h             = 1/Ngau;
  ipanel        = floor(linspace(1+0.5*h,Npan+1-0.5*h,Npan*Ngau));
  
else
    
  fprintf(1,'This option has not been implemented.\n')
  keyboard
  
end
  
return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [tt,ww] = LOCAL_get_gauss_nodes(tt_panels,Ngau)

npanels = length(tt_panels)-1;
tt   = zeros(1,npanels*Ngau);
ww   = zeros(1,npanels*Ngau);
[t_ref,w_ref] = LOCAL_lgwt(Ngau,0,1);
t_ref         = t_ref(end:(-1):1);
w_ref         = w_ref(end:(-1):1);
for i = 1:npanels
  h       = tt_panels(i+1) - tt_panels(i);
  ind     = (i-1)*Ngau + (1:Ngau);
  tt(ind) = tt_panels(i) + h*t_ref;
  ww(ind) = h*w_ref;
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [x,w] = LOCAL_lgwt(N,a,b)

%%% Use tabulated values if they exist
if (N == 5)
  TMP = [0.000000000000000,  0.568888888888889;...
         0.538469310105683,  0.478628670499366;...
         0.906179845938664,  0.236926885056189];
  xx = [TMP(end:(-1):2,1)',-TMP(:,1)'];
  ww = [TMP(end:(-1):2,2)', TMP(:,2)'];
%  xx = [-TMP(end:(-1):1,1)',TMP(2:end,1)'];
%  ww = [ TMP(end:(-1):1,2)',TMP(2:end,2)'];
elseif (N == 10)
  TMP = [0.148874338981631211,  0.295524224714752870;...
         0.433395394129247191,  0.269266719309996355;...
         0.679409568299024406,  0.219086362515982044;...
         0.865063366688984511,  0.149451349150580593;...
         0.973906528517171720,  0.066671344308688138];
  xx = [TMP(end:(-1):1,1)',-TMP(:,1)'];
  ww = [TMP(end:(-1):1,2)', TMP(:,2)'];
%  xx = [-TMP(end:(-1):1,1)',TMP(:,1)'];
%  ww = [ TMP(end:(-1):1,2)',TMP(:,2)'];
elseif (N == 20)
  TMP = [0.076526521133497333755,  0.152753387130725850698; ...
         0.227785851141645078080,  0.149172986472603746788; ...
         0.373706088715419560673,  0.142096109318382051329; ...
         0.510867001950827098004,  0.131688638449176626898; ...
         0.636053680726515025453,  0.118194531961518417312; ...
         0.746331906460150792614,  0.101930119817240435037; ...
         0.839116971822218823395,  0.083276741576704748725; ...
         0.912234428251325905868,  0.062672048334109063570; ...
         0.963971927277913791268,  0.040601429800386941331; ...
         0.993128599185094924786,  0.017614007139152118312];
  xx = [TMP(end:(-1):1,1)',-TMP(:,1)'];
  ww = [TMP(end:(-1):1,2)', TMP(:,2)'];
%  xx = [-TMP(end:(-1):1,1)',TMP(:,1)'];
%  ww = [ TMP(end:(-1):1,2)',TMP(:,2)'];
end
if exist('xx','var')
  xmid  = 0.5*(a+b);
  scale = 0.5*(b-a);
  x     = xmid + scale*xx';
  w     =        scale*ww';
  return
end

% lgwt.m
%
% This script is for computiNgau definite integrals usiNgau Legendre-Gauss 
% Quadrature. Computes the Legendre-Gauss nodes and weights  on an interval
% [a,b] with truncation order N
%
% Suppose you have a continuous function f(x) which is defined on [a,b]
% which you can evaluate at any x in [a,b]. Simply evaluate it at all of
% the values contained in the x vector to obtain a vector f. Then compute
% the definite integral usiNgau sum(f.*w);
%
% Written by Greg von Winckel - 02/25/2004
N=N-1;
N1=N+1; N2=N+2;

xu=linspace(-1,1,N1)';

% Initial guess
y=cos((2*(0:N)'+1)*pi/(2*N+2))+(0.27/N1)*sin(pi*xu*N/N2);

% Legendre-Gauss Vandermonde Matrix
L=zeros(N1,N2);

% Derivative of LGVM
Lp=zeros(N1,N2);

% Compute the zeros of the N+1 Legendre Polynomial
% usiNgau the recursion relation and the Newton-Raphson method

y0=2;

% Iterate until new points are uniformly within epsilon of old points
while max(abs(y-y0))>eps
    
    
    L(:,1)=1;
    Lp(:,1)=0;
    
    L(:,2)=y;
    Lp(:,2)=1;
    
    for k=2:N1
        L(:,k+1)=( (2*k-1)*y.*L(:,k)-(k-1)*L(:,k-1) )/k;
    end
 
    Lp=(N2)*( L(:,N1)-y.*L(:,N2) )./(1-y.^2);   
    
    y0=y;
    y=y0-L(:,N2)./Lp;
    
end

% Linear map from[-1,1] to [a,b]
x=(a*(1-y)+b*(1+y))/2;      

% Compute the weights
w=(b-a)./((1-y.^2).*Lp.^2)*(N2/N1)^2;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function u = LOCAL_pot_from_charges(xx_trg,xx_src,qq)

ntrg  = size(xx_trg,2);
nsrc  = size(xx_src,2);
dist1 = (xx_trg(1,:)' * ones(1,nsrc)) - ones(ntrg,1)*xx_src(1,:);
dist2 = (xx_trg(2,:)' * ones(1,nsrc)) - ones(ntrg,1)*xx_src(2,:);

distsq = dist1.*dist1 + dist2.*dist2;

u = -(1/(4*pi))*log(distsq)*reshape(qq,nsrc,1);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [ww,C,xx_int,xx_ext] = LOCAL_geo_elli(Npan,Ngau)

r1            = 2;    % Parameter controlling the discretization.
r2            = 1.5;  % Parameter controlling the discretization. 
t_sta         = 0;    % Start point of the discretization.
t_end         = 2*pi; % End   point of the discretization. 
tt_panels     = linspace(t_sta,t_end,Npan+1);
[tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
C             = zeros(6,length(tt));
C(1,:)        =  r1*cos(tt);
C(2,:)        = -r1*sin(tt);
C(3,:)        = -r1*cos(tt);
C(4,:)        =  r2*sin(tt);
C(5,:)        =  r2*cos(tt);
C(6,:)        = -r2*sin(tt);
ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
%h             = 1/Ngau;
%ipanel        = floor(linspace(1+0.5*h,Npan+1-0.5*h,Npan*Ngau));

nloc   = 10;
R_min  = sqrt(min(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
R_max  = sqrt(max(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
xx_int = [0.7*R_min*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          0.7*R_min*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];
xx_ext = [1.3*R_max*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          1.3*R_max*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [ww,C,xx_int,xx_ext] = LOCAL_geo_star(Npan,Ngau)

r             = 0.2;  % Parameter controlling the discretization.
k             = 5;    % Parameter controlling the discretization. 
t_sta         = 0;    % Start point of the discretization.
t_end         = 2*pi; % End   point of the discretization.
tt_panels     = linspace(t_sta,t_end,Npan+1);
[tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
C             = zeros(6,Npan*Ngau);
C(1,:)        =   1.5*cos(tt) + (r/2)*            cos((k+1)*tt) + (r/2)*            cos((k-1)*tt);
C(2,:)        = - 1.5*sin(tt) - (r/2)*(k+1)*      sin((k+1)*tt) - (r/2)*(k-1)*      sin((k-1)*tt);
C(3,:)        = - 1.5*cos(tt) - (r/2)*(k+1)*(k+1)*cos((k+1)*tt) - (r/2)*(k-1)*(k-1)*cos((k-1)*tt);
C(4,:)        =       sin(tt) + (r/2)*            sin((k+1)*tt) - (r/2)*            sin((k-1)*tt);
C(5,:)        =       cos(tt) + (r/2)*(k+1)*      cos((k+1)*tt) - (r/2)*(k-1)*      cos((k-1)*tt);
C(6,:)        = -     sin(tt) - (r/2)*(k+1)*(k+1)*sin((k+1)*tt) + (r/2)*(k-1)*(k-1)*sin((k-1)*tt);
ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
h             = 1/Ngau;
ipanel        = floor(linspace(1+0.5*h,Npan+1-0.5*h,Npan*Ngau));

nloc   = 10;
R_min  = sqrt(min(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
R_max  = sqrt(max(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
xx_int = [0.7*R_min*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          0.7*R_min*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];
xx_ext = [1.3*R_max*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          1.3*R_max*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [ww,C,xx_int,xx_ext,tt] = LOCAL_geo_tear(Npan,Ngau,Nref)

if (mod(Npan,2) == 1)
  fprintf(1,'ERROR: Npan must be even.\n')
  keyboard
end
Npan_half     = round(Npan/2);
h             = 2*pi/Npan;
tt_panels     = [ linspace(-pi,-h,Npan_half),...
                 -h*(1./(2.^(1:Nref))),...
                  0,...
                  h*(1./(2.^(Nref:(-1):1))),...
                  linspace( h, pi,Npan_half)];
[tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);

C             = zeros(6,length(tt));
C(1,:)        =      2*sign(tt).*sin(tt/2);
C(2,:)        =        sign(tt).*cos(tt/2);
C(3,:)        = -(1/2)*sign(tt).*sin(tt/2);
C(4,:)        = -tan(pi/4)*sin(tt);
C(5,:)        = -tan(pi/4)*cos(tt);
C(6,:)        =  tan(pi/4)*sin(tt);
ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
h             = 1/Ngau;
ipanel        = floor(linspace(1+0.5*h,Npan+1-0.5*h,Npan*Ngau));

nloc = 10;
R_max  = sqrt(max((C(1,:)-1).*(C(1,:)-1) + C(4,:).*C(4,:)));
xx_int = [1.5 + 0.3*(rand(1,nloc)-0.5);0.3*(rand(1,nloc)-0.5)];
xx_ext = [1 + 1.3*R_max*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
              1.3*R_max*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [ww,C,xx_int,xx_ext,tt] = LOCAL_geo_pacman(Npan,Ngau,Nref)

if (mod(Npan,2) == 1)
  fprintf(1,'ERROR: Npan must be even.\n')
  keyboard
end
Npan_half     = round(Npan/2);
h             = 2*pi/Npan;
tt_panels     = [ linspace(-pi,-h,Npan_half),...
                 -h*(1./(2.^(1:Nref))),...
                  0,...
                  h*(1./(2.^(Nref:(-1):1))),...
                  linspace( h, pi,Npan_half)];
[tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
%t_sta         = -pi;    % Start point of the discretization.
%t_end         =  pi;    % End   point of the discretization.
%tt_panels     = linspace(t_sta,t_end,Npan+1);
%[tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
C             = zeros(6,length(tt));
C(1,:)        = -     sign(tt).*sin(1.5*tt)+1;
C(2,:)        = - 1.5*sign(tt).*cos(1.5*tt);
C(3,:)        =  2.25*sign(tt).*sin(1.5*tt);
C(4,:)        =     tan(3*pi/4)*sin(tt);
C(5,:)        =     tan(3*pi/4)*cos(tt);
C(6,:)        = -   tan(3*pi/4)*sin(tt);
ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
h             = 1/Ngau;
ipanel        = floor(linspace(1+0.5*h,Npan+1-0.5*h,Npan*Ngau));

nloc = 10;
R_max  = sqrt(max((C(1,:)-1).*(C(1,:)-1) + C(4,:).*C(4,:)));
xx_int = [1.5 + 0.3*(rand(1,nloc)-0.5);0.3*(rand(1,nloc)-0.5)];
xx_ext = [1 + 1.3*R_max*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
              1.3*R_max*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];
return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [ww,C,xx_int,xx_ext] = LOCAL_geo_pacm(Npan,Ngau,Nref)

if (mod(Npan,2) == 1)
  fprintf(1,'ERROR: Npan must be even.\n')
  keyboard
end
Npan_half     = round(Npan/2);
h             = 2*pi/Npan;
tt_panels     = [ linspace(-pi,-h,Npan_half),...
                 -h*(1./(2.^(1:Nref))),...
                  0,...
                  h*(1./(2.^(Nref:(-1):1))),...
                  linspace( h, pi,Npan_half)];
              
[tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);

C             = zeros(6,length(tt));
C(1,:)        = -     sign(tt).*sin(1.5*tt)+1;
C(2,:)        = - 1.5*sign(tt).*cos(1.5*tt);
C(3,:)        =  2.25*sign(tt).*sin(1.5*tt);
C(4,:)        =     tan(3*pi/4)*sin(tt);
C(5,:)        =     tan(3*pi/4)*cos(tt);
C(6,:)        = -   tan(3*pi/4)*sin(tt);
ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
%h             = 1/Ngau;
%ipanel        = floor(linspace(1+0.5*h,Npan+1-0.5*h,Npan*Ngau));
              
nloc   = 10;
R_max  = sqrt(max((C(1,:)-1).*(C(1,:)-1) + C(4,:).*C(4,:)));
xx_int = [1.5 + 0.3*(rand(1,nloc)-0.5);0.3*(rand(1,nloc)-0.5)];
xx_ext = [1 + 1.3*R_max*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
              1.3*R_max*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [ww,C,xx_int,xx_ext] = LOCAL_geo_squa(Npan,Ngau,Nref)

if ~(mod(Npan,4) == 0)
  fprintf(1,'WARNING: Npan is not divisible by 4.\n')
  keyboard
end
Nqua = round(Npan/4);

% Construct the tt_panels objects for the strip [0,2]:
h             = 2/Nqua;
tt_pan_small     = zeros(1,Nqua + 2*Nref);
tt_pan_small(2:(Nref+1)) = h*(1./(2.^(Nref:(-1):1)));
tt_pan_small((Nref+2):(Nref+Nqua)) = linspace(h,2-h,Nqua-1);
tt_pan_small((Nref+Nqua+1):(Nqua + 2*Nref)) = 2 - h*(1./(2.^(1:Nref)));
tt_panels = [    tt_pan_small,...
             2 + tt_pan_small,...
             4 + tt_pan_small,...
             6 + tt_pan_small,...
             8];
[tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
C             = zeros(6,length(tt));
ind_e         = find((tt <= 2));
ind_n         = find((2 < tt) & (tt <= 4));
ind_w         = find((4 < tt) & (tt <= 6));
ind_s         = find((6 < tt) & (tt <= 8));
%%% Construct the "east" leg:
n_e           = length(ind_e);
C(1,ind_e)    =  ones(1,n_e);
C(2,ind_e)    = zeros(1,n_e);
C(3,ind_e)    = zeros(1,n_e);
C(4,ind_e)    = tt(ind_e)-1;
C(5,ind_e)    = ones(1,n_e);
C(6,ind_e)    = zeros(1,n_e);
%%% Construct the "north" leg:
n_n           = length(ind_n);
C(1,ind_n)    = 3 - tt(ind_n);
C(2,ind_n)    = -ones(1,n_n);
C(3,ind_n)    = zeros(1,n_n);
C(4,ind_n)    = ones(1,n_n);
C(5,ind_n)    = zeros(1,n_n);
C(6,ind_n)    = zeros(1,n_n);
%%% Construct the "west" leg:
n_w           = length(ind_w);
C(1,ind_w)    = -ones(1,n_w);
C(2,ind_w)    = zeros(1,n_w);
C(3,ind_w)    = zeros(1,n_w);
C(4,ind_w)    = 5-tt(ind_w);
C(5,ind_w)    = -ones(1,n_w);
C(6,ind_w)    = zeros(1,n_w);
%%% Construct the "south" leg:
n_s           = length(ind_s);
C(1,ind_s)    = tt(ind_s) - 7;
C(2,ind_s)    = ones(1,n_s);
C(3,ind_s)    = zeros(1,n_s);
C(4,ind_s)    = -ones(1,n_s);
C(5,ind_s)    = zeros(1,n_s);
C(6,ind_s)    = zeros(1,n_s);

ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));

nloc   = 10;
R_min  = sqrt(min(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
R_max  = sqrt(max(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
xx_int = [0.7*R_min*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          0.7*R_min*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];
xx_ext = [1.3*R_max*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          1.3*R_max*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [ww,C,xx_int,xx_ext] = LOCAL_geo_ngra(Npan,Ngau,Nref,Ngra)

if ~(mod(Npan,Ngra) == 0)
  fprintf(1,'ERROR: Each leg must have the same number of panels!\n')
  keyboard
end
Nloc = round(Npan/Ngra);

th_vec = linspace(0,2*pi,Ngra+1);
%th_vec(2:(end-1)) = th_vec(2:(end-1)) + (2*pi*0.2/Ngra)*(1-2*rand(1,Ngra-1));
r_vec  = ones(1,Ngra+1);
%r_vec(2:(end-1)) = r_vec(2:(end-1)) + 0.4*(1-2*rand(1,Ngra-1));
r_vec(2:(end-1)) = r_vec(2:(end-1)) + 0.4*((-1).^(1:(Ngra-1)));

% Construct the tt_panels objects for the strip [0,1]:
h             = 1/Nloc;
tt_pan_small     = zeros(1,Nloc + 2*Nref);
tt_pan_small(2:(Nref+1)) = h*(1./(2.^(Nref:(-1):1)));
tt_pan_small((Nref+2):(Nref+Nloc)) = linspace(h,1-h,Nloc-1);
tt_pan_small((Nref+Nloc+1):(Nloc + 2*Nref)) = 1 - h*(1./(2.^(1:Nref)));
[tt_small,ww_gauss_small] = LOCAL_get_gauss_nodes([tt_pan_small,1],Ngau);

Npart = length(tt_small);
C = zeros(6,Ngra*Npart);
tt_panels = zeros(1,Ngra*length(tt_pan_small) + 1);
tt = zeros(1,Ngra*Npart);
ww_gauss = zeros(1,Ngra*Npart);
for j = 1:Ngra
  ind_pan = (j-1)*length(tt_pan_small) + (1:length(tt_pan_small));
  tt_panels(ind_pan) = (j-1) + tt_pan_small;
  ind = (j-1)*Npart + (1:Npart);
  tt(ind) = (j-1) + tt_small;
  ww_gauss(ind) =  ww_gauss_small;
  x1sta = r_vec(j  )*cos(th_vec(j));
  x2sta = r_vec(j  )*sin(th_vec(j));
  x1end = r_vec(j+1)*cos(th_vec(j+1));
  x2end = r_vec(j+1)*sin(th_vec(j+1));
  len   = sqrt((x1sta - x1end) * (x1sta - x1end) + ...
               (x2sta - x2end) * (x2sta - x2end));
  C(1,ind) = x1sta + (x1end - x1sta) * tt_small;
  C(2,ind) = (x1end - x1sta) * ones(1,length(tt_small));
  C(3,ind) = zeros(1,length(tt_small));
  C(4,ind) = x2sta + (x2end - x2sta) * tt_small;
  C(5,ind) = (x2end - x2sta) * ones(1,length(tt_small));
  C(6,ind) = zeros(1,length(tt_small));
end
tt_panels(end) = Ngra;
    
ww     = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));

nloc   = 10;
R_min  = sqrt(min(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
R_max  = sqrt(max(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
xx_int = [0.7*R_min*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          0.7*R_min*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];
xx_ext = [1.3*R_max*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          1.3*R_max*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_get_A_diag(CE,ind)

beta = -1/(2*pi);
n    = length(ind);
ww   = CE(10,ind);
dd1  = CE(1,ind)' * ones(1,n) - ones(n,1) * CE(1,ind);
dd2  = CE(4,ind)' * ones(1,n) - ones(n,1) * CE(4,ind);
nn1  = ones(n,1)*( CE(5,ind)./sqrt(CE(2,ind).*CE(2,ind) + CE(5,ind).*CE(5,ind)));
nn2  = ones(n,1)*(-CE(2,ind)./sqrt(CE(2,ind).*CE(2,ind) + CE(5,ind).*CE(5,ind)));
ddsq  = (dd1.*dd1 + dd2.*dd2) + eye(n);

% Compute the "pure" double layer kernel (note that diagonal entries will be zero).
D     = beta*(nn1.*dd1 + nn2.*dd2)./ddsq;

% Add the diagonal term.
d_vec = -(-CE(6,ind).*CE(2,ind) + CE(3,ind).*CE(5,ind))./((4*pi)*(CE(2,ind).^2 + CE(5,ind).^2).^1.5);
D     = D + diag(d_vec);

% Add the contribution from the "correction" sources.
ddsq  = (CE(1,ind)' * ones(1,n) - ones(n,1)*CE(7,ind)).^2 + ...
        (CE(4,ind)' * ones(1,n) - ones(n,1)*CE(8,ind)).^2;
D     = D + 0.5*beta*log(ddsq); 

% Scale by the root of the quadrature weights.
A = diag(0.5*CE(9,ind)) + sqrt(ww' * ww).*D;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_get_A_offd(CE,ind1,ind2)

beta = -1/(2*pi);
n1   = length(ind1);
n2   = length(ind2);
dd1  = CE(1,ind1)' * ones(1,n2) - ones(n1,1) * CE(1,ind2);
dd2  = CE(4,ind1)' * ones(1,n2) - ones(n1,1) * CE(4,ind2);
nn1  = ones(n1,1)*( CE(5,ind2)./sqrt(CE(2,ind2).*CE(2,ind2) + CE(5,ind2).*CE(5,ind2)));
nn2  = ones(n1,1)*(-CE(2,ind2)./sqrt(CE(2,ind2).*CE(2,ind2) + CE(5,ind2).*CE(5,ind2)));
ddsq = dd1.*dd1 + dd2.*dd2;

% Compute the "pure" double layer kernel.
D    = beta*(nn1.*dd1 + nn2.*dd2)./ddsq;

% Add the contribution from the "correction" sources.
ddsq = (CE(1,ind1)' * ones(1,n2) - ones(n1,1)*CE(7,ind2)).^2 + ...
       (CE(4,ind1)' * ones(1,n2) - ones(n1,1)*CE(8,ind2)).^2;
D    = D + 0.5*beta*log(ddsq);
     
% Scale by the root of the quadrature weights.
ww1  = CE(10,ind1);
ww2  = CE(10,ind2);
A    = (sqrt(ww1'*ww2)).*D;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function u = LOCAL_eval_pot(xx_trg,CE,sigma)

C = CE(1:6,:);
ww = CE(10,:);
beta = -1/(2*pi);

n_trg = size(xx_trg,2);
n_src = size(C,2);

dd1 = xx_trg(1,:)' * ones(1,n_src) - ones(n_trg,1) * C(1,:);
dd2 = xx_trg(2,:)' * ones(1,n_src) - ones(n_trg,1) * C(4,:);
nn1 = ones(n_trg,1)*( C(5,:)./sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:)));
nn2 = ones(n_trg,1)*(-C(2,:)./sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:)));

D = -(1/(2*pi))*(nn1.*dd1 + nn2.*dd2)./(dd1.*dd1 + dd2.*dd2);

ddsq = (xx_trg(1,:)' * ones(1,n_src) - ones(n_trg,1) * CE(7,:)).^2 + ...
       (xx_trg(2,:)' * ones(1,n_src) - ones(n_trg,1) * CE(8,:)).^2;

u = (D + 0.5*beta*log(ddsq)) * (ww' .* sigma);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
